﻿#ifndef __IKD_OX_GAME_WIN_OXCP_H
#define __IKD_OX_GAME_WIN_OXCP_H


#include <stdint.h>

namespace OX {
	/**
	* @brief COMポインタ用参照カウンタ操作
	*/
	template <class T>
	class cp_counter {
	public:
		/**
		* @brief AddRef呼び出し
		*/
		static void addRef(T* ptr) { if (ptr) ptr->AddRef(); }
		/**
		* @brief Release呼び出し
		*/
		static void release(T* ptr) { if (ptr) ptr->Release(); }
		/**
		* @brief 参照数を取得
		*/
		static uint32_t ref(T* ptr) {
			if ( ptr == 0 )
				return 0;
			auto c = ptr->AddRef();
			ptr->Release();
			return c;
		}
	};

	/**
	* @brief COMポインタ
	*
	* COM（IUnknownインターフェイスを継承したクラス）の寿命を管理します。
	* コンストラクタもしくはtoCreatorメソッドで管理対象になったCOMオブジェクトの
	* AddRef及びReleaseを自動的に行います。
	* 参照数の整合性が取れなくなるため、COMポインタ管理になったCOMオブジェクトに対して
	* 使用者が明示的にAddRef及びReleaseしてはいけません。
	*/
	template <class T, class COUNTER = cp_counter<T>>
	class cp {
		T* ptr;

		void copy(cp<T> &r) {
			COUNTER::release(ptr);
			ptr = r.getPtr();
			COUNTER::addRef(ptr);
		}

		void copy( const cp<T> &r ) {
			COUNTER::release( ptr );
			ptr = (T*)r.getPtr();
			COUNTER::addRef( ptr );
		}

		template <class T2>
		void copy(cp<T2> &r) {
			COUNTER::release(ptr);
			ptr = r.getPtr();
			COUNTER::addRef(ptr);
		}

	public:
		cp() : ptr() {}

		/**
		* @brief Unknown登録
		* @param[in] src Unknown継承クラスのオブジェクトへのポインタ
		* @param[in] isAddRef 何らかの理由で内部でカウントアップしたくない場合にtrue（デフォルトのfalseのままにする事を推奨します）
		*
		* 各種生成関数で作成したUnknown継承クラスのオブジェクトを登録します。
		* srcに渡されるUnknownオブジェクトは少なくとも1カウントアップされているとします。
		*/
		cp(T* src, bool isAddRef = false) : ptr(src) {
			if (src && isAddRef)
				COUNTER::addRef(ptr);
		}

		/**
		* @brief コピー
		* @param[in] r コピー元のCOMポインタ
		*/
		cp(cp<T> &r) : ptr() { copy(r); }

		/**
		* @brief コピー（型違いコピー）
		* @param[in] r コピー元のCOMポインタ
		*/
		template <class T2>
		cp(cp<T2> &r) : ptr() { copy(r); }

		~cp() { COUNTER::release(ptr); }

		/**
		* @brief ポインタ取得
		* @return 生ポインタ
		*/
		T* getPtr() { return ptr; }

		/**
		* @brief ポインタ取得（const）
		* @return 生ポインタ
		*/
		const T* getPtr() const { return ptr; }

		/**
		* @brief ポインタのポインタを取得（const）
		* @return 生ポインタのポインタ
		*/
		T** getPtrPtr() { return &ptr; }

		/**
		* @brief アロー演算子
		*/
		T* operator ->() { return ptr; }

		/**
		* @brief 代入演算子
		*/
		cp<T> &operator =(const cp<T> &r) {
			if (r.getPtr() != ptr)
				copy(r);
			return *this;
		}

		/**
		* @brief 代入演算子
		*/
		cp<T> &operator =(const uint32_t &r) {
			COUNTER::release(ptr);
			ptr = 0;
			return *this;
		}

		/**
		* @brief 生成者へポインタへの参照を引き渡し
		*
		* COMオブジェクトを生成する関数などに内部のポインタへの参照を直接渡します。
		* すでのCOMポインタがある場合はReleaseメソッドが呼ばれ置き換わります。
		*/
		T** toCreator() {
			COUNTER::release(ptr);
			ptr = 0;
			return &ptr;
		}
		
		/**
		* @brief 参照数を取得
		*/
		uint32_t ref() {
			if ( ptr == 0 )
				return 0;
			auto cnt = ptr->AddRef();
			ptr->Release();
			return cnt - 1;
		}
	};
}

#endif
